#ifndef _SFIO_H		/* protect against multiple #includes */
#define _SFIO_H	1

/* to prevent stupid C++ stdarg.h from including stdio.h */
#if __cplusplus && !_SFSTDIO_H
#ifndef pyr
#define	pyr	1
#endif
#ifndef BSD
#define BSD	1
#endif
#endif

#ifdef __STD_C
#undef __STD_C
#endif
#ifdef __STDC__
#define	__STD_C		1
#else
#if __cplusplus
#define __STD_C		1
#endif
#endif

#ifndef _ARG_
#ifdef __STD_C
#define _ARG_(x)	x
#else
#define _ARG_(x)	()
#endif
#endif /*_ARG_*/

#ifndef Void_p
#ifdef __STD_C
#define Void_p	void*
#else
#define Void_p	char*
#endif
#endif /*Void_p*/

typedef struct _sfio_	Sfile_t, Sfio_t, SFIO;
typedef struct _sfdc_	Sfdisc_t;
typedef int		(*Sfread_f)_ARG_((Sfio_t*, Void_p, int, Sfdisc_t*));
typedef int		(*Sfwrite_f)_ARG_((Sfio_t*, const Void_p, int, Sfdisc_t*));
typedef long		(*Sfseek_f)_ARG_((Sfio_t*, long, int, Sfdisc_t*));
typedef int		(*Sfexcept_f)_ARG_((Sfio_t*, int, Sfdisc_t*));

/* discipline structure */
struct _sfdc_
{
	Sfread_f	readf;		/* read function		*/
	Sfwrite_f	writef;		/* write function		*/
	Sfseek_f	seekf;		/* seek function		*/
	Sfexcept_f	exceptf;	/* to handle exceptions		*/
	Sfdisc_t	*disc;		/* the continuing discipline	*/
};

/* a file structure */
struct _sfio_
{
	unsigned char	*next;	/* next position to read/write from	*/
	unsigned char	*endw;	/* end of write buffer			*/
	unsigned char	*endr;	/* end of read buffer			*/
	unsigned char	*endb;	/* end of buffer			*/
	struct _sfio_	*push;	/* the stream that was pushed on	*/
	unsigned short	flags;	/* type of stream			*/
	short		file;	/* file descriptor			*/
	unsigned char	*data;	/* base of data buffer			*/
	int		size;	/* buffer size				*/
#ifdef _SFIO_PRIVATE
	_SFIO_PRIVATE
#endif
};

/* various constants */
#ifndef NULL
#define NULL	0
#endif
#ifndef EOF
#define EOF	(-1)
#endif
#ifndef SEEK_SET
#define SEEK_SET	0
#define SEEK_CUR	1
#define SEEK_END	2
#endif

/* bits for various types of files */
#define	SF_READ		0000001	/* open for reading			*/
#define SF_WRITE	0000002	/* open for writing			*/
#define SF_STRING	0000004	/* a string stream			*/
#define SF_APPEND	0000010	/* associated file is in append mode	*/
#define SF_MALLOC	0000020	/* buffered space malloc-ed		*/
#define SF_LINE		0000040	/* line buffering			*/
#define SF_SHARE	0000100	/* file stream that is shared		*/
#define SF_EOF		0000200	/* eof was detected			*/
#define SF_ERROR	0000400	/* an error happened			*/
#define SF_STATIC	0001000	/* a stream that cannot be freed	*/
#define SF_IOCHECK	0002000	/* call exceptf before doing IO		*/

#define SF_FLAGS	0001177	/* PUBLIC FLAGS PASSABLE TO SFNEW()	*/
#define SF_SETS		0003163	/* flags passable to sfset()		*/

/* exception events: SF_NEW(0), SF_READ(1), SF_WRITE(2) and the below 	*/
#define SF_SEEK		3	/* seek error				*/
#define SF_CLOSE	4	/* when stream is being closed		*/
#define SF_DPUSH	5	/* when discipline is being pushed	*/
#define SF_DPOP		6	/* when discipline is being popped	*/
#define SF_DPOLL	7	/* see if stream is ready for I/O	*/

/* for stack and disciplines */
#define SF_POPSTACK	(Sfio_t*)(0)	/* pop the stream stack		*/
#define SF_POPDISC	(Sfdisc_t*)(0)	/* pop the discipline stack	*/

/* for the notify function and discipline exception */
#define SF_NEW		0	/* new stream				*/
#define SF_SETFD	-1	/* about to set the file descriptor 	*/

#define SF_BUFSIZE	8192	/* suggested default buffer size	*/
#define SF_UNBOUND	(-1)	/* unbounded buffer size		*/

#define	sfstdin		(&_Sfstdin)	/* standard input stream	*/
#define	sfstdout	(&_Sfstdout)	/* standard output stream	*/
#define	sfstderr	(&_Sfstderr)	/* standard error stream	*/

#ifdef __STD_C
#include		<stdarg.h>
#endif

#if __cplusplus
extern "C"
{
#endif

extern int		_Sfi;
extern Sfio_t		_Sfstdin, _Sfstdout, _Sfstderr;

extern Sfio_t*		sfnew _ARG_((Sfio_t*, char*, int, int, int));
extern Sfio_t*		sfopen _ARG_((Sfio_t*, const char*, const char*));
extern Sfio_t*		sfpopen _ARG_((Sfio_t*, const char*, const char*));
extern Sfio_t*		sfstack _ARG_((Sfio_t*, Sfio_t*));
extern Sfio_t*		sftmp _ARG_((int));
extern int		_sfflsbuf _ARG_((Sfio_t*, int));
extern int		_sffilbuf _ARG_((Sfio_t*, int));
extern int		sfpurge _ARG_((Sfio_t*));
extern int		sfpoll _ARG_((Sfio_t**, int, int));
extern int		sfpeek _ARG_((Sfio_t*, char**, int));
extern char*		sfreserve _ARG_((Sfio_t*, int, int));
extern int		sfsync _ARG_((Sfio_t*));
extern int		sfclrlock _ARG_((Sfio_t*));
extern char*		sfsetbuf _ARG_((Sfio_t*, char*, int));
extern Sfdisc_t*	sfdisc _ARG_((Sfio_t*,Sfdisc_t*));
extern int		sfnotify _ARG_((void(*)(Sfio_t*, int, int)));
extern int		sfset _ARG_((Sfio_t*, int, int));
extern int		sfsetfd _ARG_((Sfio_t*, int));
extern Sfio_t*		sfpool _ARG_((Sfio_t*, Sfio_t*, int));
extern int		sfread _ARG_((Sfio_t*, Void_p, int));
extern int		sfwrite _ARG_((Sfio_t*, const Void_p, int));
extern int		sfmove _ARG_((Sfio_t*, Sfio_t*, long, int));
extern int		sfclose _ARG_((Sfio_t*));
extern long		sftell _ARG_((Sfio_t*));
extern long		sfseek _ARG_((Sfio_t*, long, int));
extern int		sfllen _ARG_((long));
extern int		sfdlen _ARG_((double));
extern int		sfputr _ARG_((Sfio_t*, const char*, int));
extern char*		sfgetr _ARG_((Sfio_t*, int, int));
extern int		sfnputc _ARG_((Sfio_t*, int, int));
extern int		_sfputu _ARG_((Sfio_t*, unsigned long));
extern int		_sfputl _ARG_((Sfio_t*, long));
extern long		_sfgetl _ARG_((Sfio_t*));
extern unsigned long	_sfgetu _ARG_((Sfio_t*));
extern long		_sfgetl _ARG_((Sfio_t*));
extern int		_sfputd _ARG_((Sfio_t*, double));
extern double		sfgetd _ARG_((Sfio_t*));
extern int		sfungetc _ARG_((Sfio_t*, int));
extern char*		_sfcvt _ARG_((double, int, int*, int*, int));
extern char*		sfprints _ARG_((const char*, ...));
extern int		sfprintf _ARG_((Sfio_t*, const char*, ...));
extern int		sfsprintf _ARG_((char*, int, const char*, ...));
extern int		sfscanf _ARG_((Sfio_t*, const char*, ...));
extern int		sfsscanf _ARG_((const char*, const char*, ...));
extern int		sfvprintf _ARG_((Sfio_t*, const char*, va_list));
extern int		sfvscanf _ARG_((Sfio_t*, const char*, va_list));

/* io functions with discipline continuation */
extern int		sfrd _ARG_((Sfio_t*, Void_p, int, Sfdisc_t*));
extern int		sfwr _ARG_((Sfio_t*, const Void_p, int, Sfdisc_t*));
extern long		sfsk _ARG_((Sfio_t*, long, int, Sfdisc_t*));
extern int		sfpkrd _ARG_((int, char*, int, int, long, int));

/* function analogues of fast in-line functions */
extern int		sfgetc _ARG_((Sfio_t*));
extern long		sfgetl _ARG_((Sfio_t*));
extern unsigned long	sfgetu _ARG_((Sfio_t*));
extern int		sfputc _ARG_((Sfio_t*,int));
extern int		sfputd _ARG_((Sfio_t*,double));
extern int		sfputl _ARG_((Sfio_t*,long));
extern int		sfputu _ARG_((Sfio_t*,unsigned long));
extern int		sfslen _ARG_((void));
extern int		sfulen _ARG_((unsigned long));
extern long		sfsize _ARG_((Sfio_t*));
extern int		sfclrerr _ARG_((Sfio_t*));
extern int		sfeof _ARG_((Sfio_t*));
extern int		sferror _ARG_((Sfio_t*));
extern int		sffileno _ARG_((Sfio_t*));
extern int		sfstacked _ARG_((Sfio_t*));
extern char*		sfecvt _ARG_((double,int,int*,int*));
extern char*		sffcvt _ARG_((double,int,int*,int*));

#if __cplusplus
}
#endif

/* fast in-line functions */
#define sfputc(f,c)	((f)->next >= (f)->endw ? \
				_sfflsbuf(f,(int)((unsigned char)(c))) : \
				(int)(*(f)->next++ = (unsigned char)(c)))
#define sfgetc(f)	((f)->next >= (f)->endr ? _sffilbuf(f,0) : (int)(*(f)->next++))
#define sfslen()	(_Sfi)
#define sffileno(f)	((f)->file)
#define sfeof(f)	((f)->flags&SF_EOF)
#define sferror(f)	((f)->flags&SF_ERROR)
#define sfclrerr(f)	((f)->flags &= ~(SF_ERROR|SF_EOF))
#define sfstacked(f)	((f)->push != (Sfio_t*)0)

/* coding long integers in a portable and compact fashion */
#define SF_SBITS	6
#define SF_UBITS	7
#define SF_SIGN		(1 << SF_SBITS)
#define SF_MORE		(1 << SF_UBITS)
#define SF_U1		SF_MORE
#define SF_U2		(SF_U1*SF_U1)
#define SF_U3		(SF_U2*SF_U1)
#define SF_U4		(SF_U3*SF_U1)
#define sfulen(v)	((v) < SF_U1 ? 1 : (v) < SF_U2 ? 2 : \
			 (v) < SF_U3 ? 3 : (v) < SF_U4 ? 4 : 5)
#define sfgetu(f)	((_Sfi = sfgetc(f)) < 0 ? -1 : \
				((_Sfi&SF_MORE) ? _sfgetu(f) : (unsigned long)_Sfi))
#define sfgetl(f)	((_Sfi = sfgetc(f)) < 0 ? -1 : \
				((_Sfi&(SF_MORE|SF_SIGN)) ? _sfgetl(f) : (long)_Sfi))
#define sfputu(f,v)	_sfputu((f),(unsigned long)(v))
#define sfputl(f,v)	_sfputl((f),(long)(v))
#define sfputd(f,v)	_sfputd((f),(double)(v))

#define sfecvt(v,n,d,s)	_sfcvt((v),(n),(d),(s),1)
#define sffcvt(v,n,d,s)	_sfcvt((v),(n),(d),(s),0)

#endif /* _SFIO_H */
